/*EM algorithm for mixture distributions.  Imagine we observe
 Y[1:n] coming from a mixture of poisson distributions.  We do not
 know which of the component distributions a given Y[i] comes from.
 We would like to estimate the parameters in all of the distributions,
 as well as the relative abundances of the distributions in the population.*/

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>
#include <algorithm>
#include <math.h>
#include <boost/math/distributions.hpp>
#include <valarray>
#include <iomanip>
#include "ebwt_search.h"

using namespace std;

double POISSON(double k, double i)
{
	boost::math::poisson_distribution<> poisson(i);
	return(pdf(poisson, k));

}

double POISSON_quantile(double k, double i)
{
	boost::math::poisson_distribution<> poisson(i);
	return(quantile(complement(poisson, k)));

}



int Expectation_maximization(const vector<int> &EM_data, int frag_len)
{

	// EM-algorithm
	int D=2; // number of components in the mixture.
	int M=5000; // number of iterations of EM algorithm
	int flag=0; // Flag to stop EM algorithm as soon as the parameters converge.
    int n_tot=0;
	int number_of_tagcounts = EM_data.size();
	int tag_cutoff;
	double lambda[D][M]; // matrix to store estimated lambdas of each component at every iteration of EM algorithm.
	double p[D][M]; // matrix to store estimated p_values of each component at every iteration of EM algorithm.
	double Den[D][number_of_tagcounts]; // matrix to store estimated membership probabilites for each data point.
	double Obj[M];
    double lambda_nonpeak;
    uint64_t GenomeSize;
    if (genome == "hs") GenomeSize=2700000000;
    if (genome == "mm") GenomeSize=1870000000;
    if (genome == "ce") GenomeSize=90000000;
    if (genome == "dm") GenomeSize=120000000;
    //cout <<"GenomeSize\t"<<GenomeSize;
    lambda[1][0]=double(frag_len)*NUM_OF_READS/GenomeSize;
	lambda[0][0]=POISSON_quantile(10e-5, lambda[1][0]);
	p[0][0]=0.005;
	p[1][0]=0.995;
	tag_cutoff=POISSON_quantile(10e-50, lambda[1][0]);
    vector<double> Objective;
	for(int i=0; i<number_of_tagcounts; i++)
	{
		if(i>tag_cutoff)
		{
			break;
			number_of_tagcounts=i;
		}
		n_tot=n_tot+EM_data[i];

	}

	for(int k=0; k<=M-2;k++)
		{
			if(flag==1)
			{
				break;
			}
			for(int r=0; r<=D-1;r++ )
			{
				std::valarray<double> Num_lambda(number_of_tagcounts);
				std::valarray<double> Din_lambda(number_of_tagcounts);

				for(int j=0; j<=tag_cutoff;j++ )
				{
					Den[r][j]=p[r][k]*(POISSON(j,lambda[r][k]))/(p[0][k]*(POISSON(j,lambda[0][k]))+p[1][k]*(POISSON(j,lambda[1][k])));
					Num_lambda[j]=j*EM_data[j]*Den[r][j];
					Din_lambda[j]=EM_data[j]*Den[r][j];
				}
				double Tot_Num_lambda= Num_lambda.sum();
				double Tot_Din_lambda= Din_lambda.sum();
               			lambda[r][k+1]=Tot_Num_lambda/Tot_Din_lambda;
               			p[r][k+1]=Tot_Din_lambda/n_tot;
                        lambda_nonpeak=lambda[0][k+1];
			}

                if(abs(lambda[0][k+1]-lambda[0][k])<0.00001 && abs(p[0][k+1]-p[0][k])<0.00001 && abs(lambda[1][k+1]-lambda[1][k])<0.00001 && abs(p[1][k+1]-p[1][k])<0.00001)
				{
					//cout<<"# of EM iterations"<<"\t"<<k<<endl;
					//cout<<"est_lambda_peak"<<"\t"<<lambda[0][k+1]<<"\t"<<"est_p_value_peak"<<"\t"<<p[0][k+1]<<endl;
					//cout<<"est_lambda_nonpeak"<<"\t"<<lambda[1][k+1]<<"\t"<<"est_p_value_nonpeak"<<"\t"<<p[1][k+1]<<endl;
					flag=1;
				}

            double Value=0;
            for(int j=0; j<=tag_cutoff; j++)
            {
             Value=Value+EM_data[j]*log10(p[0][k]*POISSON(j,lambda[0][k])+p[1][k]*POISSON(j,lambda[1][k]));
            }
            Objective.push_back(Value);
            /*if(k>2 && abs(Objective[k]-Objective[k-1])<0.0000001){
            break;
            }*/
            //cout<<"iteration\t"<<k<<"\tLogLikelihood\t"<<std::setprecision(20)<<Objective[k]<<endl;
		}

		//cout<<endl<<"***Membership probabilities***"<<endl<<endl;
		//cout<<"Tag count"<<"\t"<<"component-1"<<"\t"<<"component-2"<<endl<<endl;
		for(int i=0;i<=tag_cutoff; i++)
            {

                //cout<<i<<"\t"<<std::fixed <<setprecision(2)<<float(Den[0][i])<<"\t"<<float(Den[1][i])<<endl;

            }

	for(int i=0;i<=tag_cutoff; i++)
	{

            //cout<<"test1\t"<<Den[0][i]*100<<"\t"<<emCutoff*100<<endl;
            if(abs(Den[0][i]-emCutoff)*100<=1 || (Den[0][i]>=emCutoff))
            {
                //cout<<"test\t"<<Den[0][i]<<"\t"<<emCutoff<<endl;
                 if(emCutoff-Den[0][i-1]<=Den[0][i]-emCutoff)
                {
               return i-1;
                }
                else{
                 return i;
               }
            }


	}
}
